#!/bin/bash

#
# Parallel design sweeping script for Design Compiler. It can be used to sweep
# both clock constraints and VHDL design generics simultaneously. The tool is
# used by altering the global sweeping settings, found just below this note, and
# the sweep results are put in a newly created directory named
# 'sweep_<design_name>_<date_of_sweep>'. All library technology settings that
# are to be shared among sweeps are to be put in the local configuration file
# '.synopsys_dc.setup' which is sourced by all instances of Design Compiler.
# Use this script at your own risk or whatever...
#
# Author: Mikael Henriksson
#


#
# Global sweeping settings begin:
# -------------------------------
#

# Top level design entity name.
DESIGN_NAME="my_divider"

# BASH-style array with design files to be analyzed by Design Compiler. Analysis
# of files are done from left (first element) to right (last element).
DESIGN_FILES=("div.vhdl")

# BASH-style array with periods (in the technology time unit) to be sweept.
PERIODS=(10.0 5.0 3.333 2.5 2.0 1.666 1.428 1.25 1.111 1.0)

# BASH-style array with generics to be sweept. Leave empty if no design generic
# should be set. Multiple generics can be assigned for one sweep by comma
# separating each generic assignment in the BASH-array element.
GENERICS=( "float_ip=0,pipe_stages=1" "float_ip=0,pipe_stages=2"
           "float_ip=1,pipe_stages=1" "float_ip=1,pipe_stages=2" )

# Design clock name to be sweept. The design MUST have a clock signal.
CLOCK_NAME="clk"

# Enable Design Compiler adaptive retiming through 'set_optimize_registers' for
# ${DESIGN_NAME}, and use 'compile_ultra -retime' to compile the design. This
# option is applied to all designs sweept. Set to 'YES' or 'NO' (note the 
# CAPITALIZATION).
RETIME_ENABLE="YES"

# Exclude resulting sweep poits from 'result.csv' if they have negative slack.
# If set to 'NO', sweep points with negative slack are kept in the resulting
# CSV file, but are still distinguishible through the slack column, which will
# be negative. Set to 'YES' or 'NO' (note the CAPITALIZATION).
EXCLUDE_NEGATIVE_SLACK="NO"

# Export a gate-level netlist from each synthesised designs. If set to 'YES',
# three files will be generated; 1) a Verilog gate-level netlist
# '${DESIGN_NAME}-post-synth.v', 2) an SDC design constraint file
# '${DESIGN_NAME}-post-synth.sdc', and 3) a Synopsys combinded netlist and
# constraint .DDC file '${DESIGN_NAME}-post-synth.ddc'. Set to 'YES' or 'NO'
# (note the CAPITALIZATION)
EXPORT_NETLIST="YES"

# Maximum number parallel instances of Design Compiler that is allowed to run at
# a time. Each instance of Design Compiler will be used for synthesis of a
# single sweep point. Each instance of Design Compiler will claim one available
# Design Compiler licesnse, or wait for a license to become available (possibly,
# see variable WAIT_FOR_LICENSE below). This should idealy be set to the number
# of available Design Compiler Licenses, or lower.
MAX_CONCURRENCY=12

# Specifies the maximum number of CPU cores that each instance of Design
# Compiler is allowed to use. More cores (usually) equals faster synthesis,
# however remember that MAX_CONCURRENCY*MAX_CORES will be an approximate
# upper bound on the work load for the host machine.
MAX_CORES=4

# Wait for appropriate Design Compiler licenses when the required licenses are
# currently already in use by other Design Compiler instances. This enables
# Design Compiler License Queuing. Setting this option to 'NO' can make you
# miss sweep points if all available licenses are already in use by other Design
# Compiler instances. Set to either 'YES' or 'NO' (note the CAPITALIZATION).
WAIT_FOR_LICENSE="YES"

# Synthesis template file sent to each Design Compiler instance. Design Compiler
# will use the exact recepie from this file during all synthesis runs. It is in
# this file that custom synthesis flows can be specified, such as hiearcical
# bottom-up design flows. The provided 'template.tcl' recepie will run an
# unattended top-down synthesis.
SYNTH_TCL_TEMPLATE="template.tcl"

# String that will be appended to the top sweep directory name. Could, for
# example, be used to append technology, corrner or other vital info to the
# sweep directory name. Leave blank if unused.
TOP_DIR_APPEND=""

# Resulting units. These will be used for the generated CSV files. They should
# be set in accordance with the technology units.
PERIOD_DIM="ns"
POWER_DIM="mW"
AREA_DIM="um^2"

#
# -----------------------
# End of global settings.
#


# Generate a CSV sweep result file from a sweep result. Only the sweeps whose
# slack where not violated are extracted.
function sweep_to_csv() {
    local sweep_data_dir="${1}"
    local csv_out="${2}"

    local timing_files
    if [ "${EXCLUDE_NEGATIVE_SLACK}" = "YES" ]; then
        # Extract non-slack violated 'timing.txt' files. Files with violated
        # slack should contain the term "(VIOLATED", not necessary with ending
        # parentheses.
        timing_files=$(grep -riL "(VIOLATED" "${sweep_data_dir}"/*/timing.txt)
    else # ${EXCLUDE_NEGATIVE_SLACK} -eq "NO"
        # Extract all all 'timing.txt' files.
        timing_files=$(echo "${sweep_data_dir}"/*/timing.txt)
    fi
    local periods=$(perl -C -lne 'print for /(\d+\.\d+?)\/timing\.txt/ig' <<< \
              "${timing_files}" | sort -rh)

    # Create CSV file and write CSV header.
    touch "${csv_out}"
    echo "Period (${PERIOD_DIM}), Total cell area (${AREA_DIM})," \
         "Power (${POWER_DIM}), Slack (${PERIOD_DIM})" > "${csv_out}"

    # Iterate over successful synthesised results, which had it's timing
    # constraints met.
    while IFS='' read -r period; do
        # Extract data.
        cur_dir="${sweep_data_dir}/${period}"
        power=$(awk '/Total\s+[0-9\.e-]+\s[a-z]{0,1}W/{print $(NF-1)}' \
            "${cur_dir}/power.txt")
        area=$(awk '/Total cell area:/{print $NF}' "${cur_dir}/area.txt")
        # "(MET" and "(VIOLATED" does not necessary contain parentheses ending
        slack=$(awk '/slack (\(MET)|(\(VIOLATED)/{print $(NF)}' "${cur_dir}/timing.txt")

        # Append to CSV.
        echo "${period}, ${area}, ${power}, ${slack}" >> "${csv_out}"
    done <<< "${periods}"
}

# Join elements of BASH array, or white space separated string, in $2 by the
# string in $1.
function join_by {
    local d=$1; shift; local f=$1; shift; printf %s "$f" "${@/#/$d}";
}

# Function for generating the synthesis script to be run by Design Compiler. The
# third parameter to this function 'design_generics' is optional.
function generate_synthesis_script() {
    local period="$1"
    local sub_dir="$2"
    local design_generics="$3"

    # Use template and insert propriate settings.
    # shellcheck disable=SC2002
    cat "${SYNTH_TCL_TEMPLATE}" |\
      sed "s/__MAX_CORES__/${MAX_CORES}/" |\
      sed "s#__HDL_FILES__#$(join_by "\ " "${DESIGN_FILES_ABS[@]}")#" |\
      sed "s/__PERIOD__/${period}/" |\
      sed "s/__CLOCK_NAME__/${CLOCK_NAME}/" |\
      sed "s/__DESIGN_GENERICS__/${design_generics}/" |\
      sed "s/__DESIGN_NAME__/${DESIGN_NAME}/" |\
      sed "s/__RETIME_ENABLE__/${RETIME_ENABLE}/" |\
      sed "s/__EXPORT_NETLIST__/${EXPORT_NETLIST}/" >\
        "$sub_dir/sim.tcl"
}

# Function used for running Design Compiler with the different designs. The
# design script will export this function, fork of new bash-instances with xargs
# and invoke this function. Note to programmer, that means this function cannot
# access any unexported items (variables, functions, etc).
function run_design_compiler() {

    # Change active directory to design directory.
    local design_directory="$1"
    cd "$design_directory" || exit 1

    # Get time of sweep start.
    start_time="$(date +"%H:%M:%S")"

    # Activate Design Compiler License Queuing if requiered in global settings.
    if [ ${WAIT_FOR_LICENSE} = "YES" ]; then
        export SNPSLMD_QUEUE="true"
    fi

    # Source the automatically generated design options file. This file contains
    # 'period' and 'design'generics'.
    # shellcheck disable=SC1091
    source .design_options.sh

    # Display sweep start.
    if [ "$design_generics" = "" ]; then
        echo "[ PID: $$ ]: $start_time, starting synthesis: '$DESIGN_NAME'" \
             "@ $period $PERIOD_DIM."
    else
        echo "[ PID: $$ ]: $start_time, starting synthesis: '$DESIGN_NAME'" \
             "[$design_generics] @ $period $PERIOD_DIM."
    fi

    #
    # Run design compiler with the auto generated TCL build script. All logging
    # is done in the local 'log.txt' file. Custom Design Compiler exit codes:
    #   0: Success
    #   1: Analyze failure
    #   2: Elaborate failure
    #   3: Create clock failure
    #   3: Compile failure
    dc_error_type=("Success" "analyze" "elaborate" "create_clock" "compile")
    dc_shell -f sim.tcl 1>log.txt 2>&1
    dc_exit_code="$?"
    if [ "$dc_exit_code" -ne 0 ]; then
        current_time="$(date +"%H:%M:%S")"
        echo "[ PID: $$ ]: $current_time, Design Compiler non-zero exit code:" \
            "$dc_exit_code (abnormal '${dc_error_type[$dc_exit_code]}'-phase)."
    fi
}

# Signal SIGINT and SIGTERM interrupt handlers. Both are registered once the
# setup is complete and the actual sweep has started.
function sigint_handler() {
    echo "${0}: SIGINT ignored. Send SIGTERM to terminate sweep."
    echo "[ INFO: ] Sweep script PID: ${$}. Signal SIGTERM can be sent to" \
         "sweep script with 'kill':"
    echo "[ INFO: ] $> kill -s SIGTERM ${$}"
    echo "[ INFO: ] from a free terminal."
}
function sigterm_handler() {
    echo "${0}: SIGTERM recieved, propagating to workers and terminating."
    kill -s SIGTERM -"${1}"
    exit 1
}

# Export functions and variables needed by the forked workers.
export -f run_design_compiler
export DESIGN_NAME
export PERIOD_DIM
export WAIT_FOR_LICENSE
export RETIME_ENABLE
export EXPORT_NETLIST

# Sanity checks for the global settings.
if [ -z "$CLOCK_NAME" ]; then
    echo "Error: No clock name (\$CLOCK_NAME) provided."
    exit 1
fi
if [ ${#PERIODS[@]} -lt 1 ]; then
    echo "Error: Less than one clock constraint (\$PERIODS)."
    exit 1
fi
if [ "$MAX_CONCURRENCY" -lt 1 ]; then
    echo "Error: Maximum concurrency (\$MAX_CONCURRENCY) < 1."
    exit 1
fi
if [ "${#DESIGN_FILES[@]}" -lt 1 ]; then
    echo "Error: No design files (\$DESIGN_FILES) provided."
    exit 1
fi
for file in "${DESIGN_FILES[@]}"; do
    if ! [ -f "$file" ]; then
        echo "Error: Could not locate design file '$file'."
        exit 1
    fi
done
if ! [[ "${EXCLUDE_NEGATIVE_SLACK}" =~ ^(YES)|(NO)$ ]]; then
    echo "Error: \${EXCLUDE_NEGATIVE_SLACK} not in { 'YES', 'NO' };"
fi
if ! [[ "$RETIME_ENABLE" =~ ^(YES)|(NO)$ ]]; then
    echo "Error: \$RETIME_ENABLE not in { 'YES', 'NO' }."
    exit 1
fi
if ! [[ "${WAIT_FOR_LICENSE}" =~ ^(YES)|(NO)$ ]]; then
    echo "Error: \$WAIT_FOR_LICENSE not in { 'YES', 'NO' }."
    exit 1
fi
if ! [[ "${EXPORT_NETLIST}" =~ ^(YES)|(NO)$ ]]; then
    echo "Error: \$WAIT_FOR_LICENSE not in { 'YES', 'NO' }."
    exit 1
fi

# Assure all periods are in numeric format, with max one decimal delimiter.
for period in "${PERIODS[@]}"; do
    if ! [[ "$period" =~ (^0$)|(^0\.[0-9]+$)|(^[1-9][0-9]*$)|(^[1-9][0-9]*\.[0-9]+$) ]]; then
        echo "Period '$period' in \$PERIODS is not in numeric format."
        exit 1
    fi
done

# Make sure template TCL-file exists.
if [ ! -f "${SYNTH_TCL_TEMPLATE}" ]; then
    echo "Error: Could not locate template script file '${SYNTH_TCL_TEMPLATE}'."
    exit 1
fi

# Make sure dc_shell is accessable through PATH.
which dc_shell 1>/dev/null 2>&1 && HAS_DC=1 || HAS_DC=0
if [ "$HAS_DC" = "0" ]; then
    echo "Error: Design Compiler (dc_shell) not available through \$PATH."
    exit 1
fi

# If the '.synopsys_dc.setup' setup file, which should contain technology
# settings, is missing, query the user and ask if they would like to continue.
if [ ! -f ".synopsys_dc.setup" ]; then
    read -r -p "Could not locate file '.synopsys_dc.setup'. Are you sure you \
want to continure? [y/N] " response
    if [[ "$response" =~ ^([yY][eE][sS]|[yY])$ ]]; then
        echo "Continuing without '.synopsys_dc.setup'."
    else
        echo "Aborting."
        exit 1
    fi
fi

# Get absolute path of design files, which is necessary for the sweep.
DESIGN_FILES_ABS=()
for file in "${DESIGN_FILES[@]}"; do
    DESIGN_FILES_ABS+=("$(realpath "$file")")
done

# Get script launch date.
SCRIPT_LAUNCH_DATE=$(date +"%F_%H:%M:%S")


#
# Setup of sweep complete. Generate the sweep top directory, launch all sweep
# workers according to the max concurreny and register signal handlers.
#

# Create sweep top directory and copy this sweep script to it. Copying this
# sweep script lets the user revisit the settings used for a sweep.
if [ -z "$TOP_DIR_APPEND" ]; then
    TOP_DIR="sweep_${DESIGN_NAME}_${SCRIPT_LAUNCH_DATE}"
else
    TOP_DIR="sweep_${DESIGN_NAME}_${TOP_DIR_APPEND}_${SCRIPT_LAUNCH_DATE}"
fi
if ! mkdir "$TOP_DIR"; then
    echo "Error: failed to create top directory (mkdir \"${TOP_DIR}\")"
    exit 1
fi
echo "[ INFO: ] Created top directory: '${TOP_DIR}'"
cp "$(realpath "${0}")" "${TOP_DIR}"
echo "[ INFO: ] Copied ${0} into top directory"
echo "" # Line break

# Initialize designs. We create one sub directory for each design that is
# to be sweept. A 'design' is exactly one setting from ${DESIGN_GENERICS[@]}.
# Further, we substitue design options into '${SYNTH_TCL_TEMPLATE}' and copy the
# resulting synthesis script into the design directory.
SYNTH_DIRS=""
if [ "${#GENERICS[@]}" -gt 0 ]; then

    echo "Using '${DESIGN_FILES[*]}' to create design(s):"
    for design_generics in "${GENERICS[@]}"; do
        echo " [*] '$DESIGN_NAME' with generic setting '$design_generics':"
        period_cnt=1
        for period in "${PERIODS[@]}"; do

            # Create design and period subdirectory.
            sub_dir="$TOP_DIR/$design_generics/$period"
            mkdir -p "$sub_dir"
            if [ "$period_cnt" -ne "${#PERIODS[@]}" ]; then
                echo "  ├── Period: $period $PERIOD_DIM."
            else
                echo "  └── Period: $period $PERIOD_DIM."
            fi

            # Generate current design and period synthesis script from template.
            generate_synthesis_script "$period" "$sub_dir" "$design_generics"

            # Possibly copy the Design Compiler global settings/technology file.
            if [ -f ".synopsys_dc.setup" ]; then
                cp ".synopsys_dc.setup" "$sub_dir/"
            fi

            # Append new synthesis directory to sweep list.
            SYNTH_DIRS="${sub_dir}|${SYNTH_DIRS}"
            ((period_cnt=period_cnt+1))

            # Add period and design options to design option file.
            design_options="${sub_dir}/.design_options.sh"
            touch "$design_options"
            echo "period=\"$period\"" >> "$design_options"
            echo "design_generics=\"$design_generics\"" >> "$design_options"
        done
    done

else # "${GENERICS[@]}" is empty.

    echo "Using '${DESIGN_FILES[*]}' to create design:"
    echo " [*] '$DESIGN_NAME':"
    period_cnt=1
    for period in "${PERIODS[@]}"; do

        # Create period subdirectory.
        sub_dir="$TOP_DIR/$period"
        mkdir -p "$sub_dir"
        if [ "$period_cnt" -ne "${#PERIODS[@]}" ]; then
            echo "  ├── Period: $period $PERIOD_DIM."
        else
            echo "  └── Period: $period $PERIOD_DIM."
        fi

        # Generate current design and period synthesis script from template.
        generate_synthesis_script "$period" "$sub_dir"

        # Possibly copy the Design Compiler global settings/technology file.
        if [ -f ".synopsys_dc.setup" ]; then
            cp ".synopsys_dc.setup" "$sub_dir/"
        fi

        # Append new synthesis directory to sweep list.
        SYNTH_DIRS="${sub_dir}|${SYNTH_DIRS}"
        ((period_cnt=period_cnt+1))

        # Add period and design options to design option file.
        design_options="${sub_dir}/.design_options.sh"
        touch "$design_options"
        echo "period=\"$period\"" >> "$design_options"
    done
fi

# Display start of process pool log message.
printf "\nParallel synthesis log (max $MAX_CONCURRENCY concurrent processes):\n"

# Trap incoming SIGINT signals, which might be accidental and can prematurely
# terminate the sweep. User should use SIGTERM to terminate sweep beyond this
# point.
trap 'sigint_handler' SIGINT

# Launch workers according to the max concurrency with XARGS. By launching XARGS
# through SETSID, all workers are launched in the same process group, and more
# importantly, in a new process group different from that of this parent sweep
# script. This shileds all worker processes from recieving SIGINT on accidental
# shell <Ctrl+C>. The process group id (PGID) of all workers is equal to the
# process id (PID) of XARGS, and can be used if a signal must be sent to all
# workers.
printf '%s' "${SYNTH_DIRS%?}" | setsid xargs -L1 -d "|" --max-procs="$MAX_CONCURRENCY"\
    bash -c 'run_design_compiler $@' _ &
XARGS_PID="$!"

# Register SIGTERM handler.
trap "sigterm_handler ${XARGS_PID}" SIGTERM

# Wait for all sweep workers to finish.
while kill -s 0 ${XARGS_PID} 1>/dev/null 2>&1; do
    wait ${XARGS_PID}
done

# Display end of process pool log message.
echo "End of parallel synthesis log."


#
# Generate CSV files from sweep result.
#
if [ "${#GENERICS[@]}" -eq 0 ]; then
    # Sweep without design generics.
    work_dir="$TOP_DIR"
    output_csv_file="$work_dir/result.csv"
    echo "Compiling CSV results into '$output_csv_file'."
    sweep_to_csv "$work_dir" "$output_csv_file"
else
    # Sweep with design generics.
    for design_generics in "${GENERICS[@]}"; do
        work_dir="$TOP_DIR/$design_generics"
        output_csv_file="$work_dir/result.csv"
        sweep_to_csv "$work_dir" "$output_csv_file"
    done

    # Compile all CSV files into one top CSV result file.
    output_csv_file="$TOP_DIR/result.csv"
    echo "Compiling CSV results into '$output_csv_file'."
    touch "$output_csv_file"

    # Generate top CSV header.
    echo "Design, Period (${PERIOD_DIM}), Total cell area (${AREA_DIM})," \
         "Power (${POWER_DIM}), Slack (${PERIOD_DIM})" > "$output_csv_file"

    # Append results from design sweeps.
    for design_generics in "${GENERICS[@]}"; do
        WORK_DIR="$TOP_DIR/$design_generics"
        tail --lines=+2 "$WORK_DIR/result.csv" | \
            sed "s/^\(.*\)/$design_generics, \1/" >> "$output_csv_file"
    done
fi

# Script finished.
echo "Sweep finished."

